package PScm::Expr;

use strict;
use warnings;
use base qw(PScm::Token);

sub isTrue {
    my ($self) = @_;
    scalar($self->value);
}

sub is_unquote { 0 }

sub Eval {
    my ($self, $env, $cont) = @_;
    $cont->Cont($self);
}

sub strings {
    my ($self) = @_;
    return ('.', $self->as_string);
}

sub first {
    my ($self) = @_;
    die "cannot take car of ", ref($self), "\n";
}

sub rest {
    my ($self) = @_;
    die "cannot take cdr of ", ref($self), "\n";
}

sub value { $_[0] }

sub map_eval {
    my ($self, $env, $cont) = @_;
    $self->Eval($env, $cont);
}

sub Quote {
    my ($self, $env, $cont) = @_;
    $cont->Cont($self);
}

sub quote_rest {
    my ($self, $env, $cont) = @_;
    $cont->Cont($self);
}

sub is_symbol { 0 }
sub is_pair { 0 }
sub is_null { 0 }
sub is_expr { 1 }


#########################
package PScm::Expr::Atom;
use base qw(PScm::Expr);

sub new {
    my ($class, $value) = @_;
    bless \$value, $class;
}

sub value { ${ $_[0] } }

sub as_string { $_[0]->value }

#########################
package PScm::Expr::List;
use base qw(PScm::Expr);
use PScm::Continuation;

sub new {
    my ($class, @list) = @_;

    $class = ref($class) || $class;
    if (@list) {
        my $first = shift @list;
        $class->Cons($first, $class->new(@list));
    } else {
        new PScm::Expr::List::Null();
    }
}

sub Cons {
    my ($class, $first, $rest) = @_;
    return PScm::Expr::List::Pair->new($first, $rest);
}

sub as_string {
    my ($self) = @_;
    return '(' . join(' ', $self->strings) . ')';
}

###############################
package PScm::Expr::List::Pair;
use base qw(PScm::Expr::List);

use PScm::Continuation;

use constant {
    FIRST => 0,
    REST => 1,
};

sub new {
    my ($class, $first, $rest) = @_;
    bless [$first, $rest], $class;
}

sub value {
    my ($self) = @_;
    my @value = ($self->[FIRST], $self->[REST]->value);
    return @value;
}

sub first { $_[0][FIRST] }

sub rest { $_[0][REST] }

sub strings {
    my ($self) = @_;
    return ($self->[FIRST]->as_string,
            $self->[REST]->strings);
}

sub Eval {
    my ($self, $env, $cont) = @_;
    return cont {
        $self->[FIRST]->Eval(
            $env,
            cont {
                my ($op) = @_;
                $op->Apply($self->[REST], $env, $cont);
            }
        );
    };
}

sub map_eval {
    my ($self, $env, $cont) = @_;

    $self->[FIRST]->Eval(
        $env,
        cont {
            my ($evaluated_first) = @_;
            $self->[REST]->map_eval(
                $env,
                cont {
                    my ($evaluated_rest) = @_;
                    $cont->Cont($self->Cons($evaluated_first,
                                            $evaluated_rest));
                }
            );
        }
    );
}

sub is_pair { 1 }

sub Quote {
    my ($self, $env, $cont) = @_;
    if ($self->[FIRST]->is_unquote) {
        $self->[REST]->first->Eval($env, $cont);
    } else {
        $self->quote_rest($env, $cont);
    }
}

sub quote_rest {
    my ($self, $env, $cont) = @_;
    $self->[FIRST]->Quote(
        $env,
        cont {
            my ($quoted_first) = @_;
            $self->[REST]->quote_rest(
                $env,
                cont {
                    my ($quoted_rest) = @_;
                    $cont->Cont(
                        $self->Cons($quoted_first, $quoted_rest));
                }
            );
        }
    );
}

###############################
package PScm::Expr::List::Null;
use base qw(PScm::Expr::List);

sub new {
    my ($class) = @_;

    $class = ref($class) || $class;
    bless {}, $class;
}

sub value { (); }

sub strings { (); }

sub first { $_[0] }

sub rest { $_[0] }

sub is_null { 1 }

###########################
package PScm::Expr::Symbol;
use base qw(PScm::Expr::Atom);

sub Eval {
    my ($self, $env, $cont) = @_;
    $cont->Cont($env->LookUp($self));
}

sub is_unquote {
    my ($self) = @_;
    return $self->value eq "unquote";
}

sub is_symbol { 1 }

############################
package PScm::Expr::Literal;
use base qw(PScm::Expr::Atom);

###########################
package PScm::Expr::Number;
use base qw(PScm::Expr::Literal);

use Math::BigInt;

sub new {
    my ($class, $value) = @_;
    $value = new Math::BigInt($value) unless ref($value);
    $class->SUPER::new($value);
}

###########################
package PScm::Expr::String;
use base qw(PScm::Expr::Literal);

sub as_string {
    my ($self) = @_;

    my $copy = $self->value;
    $copy =~ s/\\/\\\\/sg;
    $copy =~ s/"/\\"/sg;
    return qq'"$copy"';
}

1;
